import asyncio
from typing import AsyncIterator, Awaitable, Callable, Dict, Type, Any

from pydantic import Field, field_validator
from rath.links.base import AsyncTerminatingLink
from rath.operation import GraphQLResult, Operation
from graphql import FieldNode, OperationType


def target_from_node(node: FieldNode) -> str:
    """Extract the target aka. the aliased name from a FieldNode."""
    return (
        node.alias.value if hasattr(node, "alias") and node.alias else node.name.value
    )


class AsyncMockResolver:
    """A Mock Resolver

    This is a mock resolver that can be used to mock a GraphQL server.
    Every attribute that starts with resolve_ will be used as a resolver
    for the corresponding field.
    """

    def to_dict(self) -> Dict[str, Callable[[Operation], Awaitable[Dict]]]:
        """Convert the Mock Resolver to a dict of resolvers

        Returns
        -------
        Dict[str, Callable[[Operation], Awaitable[Dict]]]
            The dict of resolvers
        """
        methods = [
            i for i in self.__class__.__dict__.keys() if i.startswith("resolve_")
        ]
        return {i[8:]: getattr(self, i) for i in methods}


ResolverDict = Dict[str, Callable[[Operation], Awaitable[Dict]]]


class AsyncMockLink(AsyncTerminatingLink):
    """A Mocklink

    This is a mocklink that can be used to mock a GraphQL server.
    You need to pass resolvers to the constructor.
    """

    query_resolver: Dict[str, Callable[[Operation], Awaitable[Dict]]] = Field(
        default_factory=dict, exclude=True
    )
    mutation_resolver: Dict[str, Callable[[Operation], Awaitable[Dict]]] = Field(
        default_factory=dict, exclude=True
    )
    subscription_resolver: Dict[str, Callable[[Operation], AsyncIterator[Dict]]] = (
        Field(default_factory=dict, exclude=True)
    )
    resolver: ResolverDict = Field(default_factory=dict, exclude=True)

    @field_validator(
        "query_resolver",
        "mutation_resolver",
        "subscription_resolver",
        "resolver",
        mode="before",
    )
    @classmethod
    def coerce_resolver(cls: Type["AsyncMockLink"], v: Any) -> ResolverDict:
        """A validator that coerces AsyncMockResolver to a dict of resolvers"""
        if isinstance(v, AsyncMockResolver):
            return v.to_dict()
        return v

    async def aexecute(self, operation: Operation) -> AsyncIterator[GraphQLResult]:
        """Executes an operation against the link

        Parameters
        ----------
        operation : Operation
            The operation to execute

        Yields
        ------
        GraphQLResult
            The result of the operation
        """

        if operation.node.operation == OperationType.QUERY:
            futures = []

            for op in operation.node.selection_set.selections:
                if isinstance(op, FieldNode):
                    if op.name.value in self.query_resolver:
                        futures.append(self.query_resolver[op.name.value](operation))
                    elif op.name.value in self.resolver:
                        futures.append(self.resolver[op.name.value](operation))
                    else:
                        raise NotImplementedError(
                            f"Mocked Resolver for Query '{op.name.value}' not in resolvers: {self.query_resolver}, {self.resolver}  for AsyncMockLink"
                        )

            resolved = await asyncio.gather(*futures)
            yield GraphQLResult(
                data={
                    target_from_node(op): resolved[i]
                    for i, op in enumerate(operation.node.selection_set.selections)
                    if isinstance(op, FieldNode)
                }
            )

        if operation.node.operation == OperationType.MUTATION:
            futures = []

            for op in operation.node.selection_set.selections:
                if isinstance(op, FieldNode):
                    if op.name.value in self.mutation_resolver:
                        futures.append(self.mutation_resolver[op.name.value](operation))
                    elif op.name.value in self.resolver:
                        futures.append(self.resolver[op.name.value](operation))
                    else:
                        raise NotImplementedError(
                            f"Mocked Resolver for Query '{op.name.value}' not in resolvers:"
                            f"{self.mutation_resolver}, {self.resolver}  for AsyncMockLink"
                        )

            resolved = await asyncio.gather(*futures)
            yield GraphQLResult(
                data={
                    target_from_node(op): resolved[i]
                    for i, op in enumerate(operation.node.selection_set.selections)
                    if isinstance(op, FieldNode)
                }
            )

        if operation.node.operation == OperationType.SUBSCRIPTION:
            futures = []
            assert (
                len(operation.node.selection_set.selections) == 1
            ), "Only one Subscription at a time possible"

            op = operation.node.selection_set.selections[0]
            if isinstance(op, FieldNode):
                if op.name.value in self.subscription_resolver:
                    iterator = self.subscription_resolver[op.name.value](operation)

                    async for event in iterator:
                        if isinstance(op, FieldNode):
                            yield GraphQLResult(data={target_from_node(op): event})
                else:
                    raise NotImplementedError(
                        f"Mocked Resolver for Query '{op.name.value}' not in resolvers"
                        f": {self.subscription_resolver}, {self.resolver}  for AsyncMockLink"
                    )

            async for event in iterator:
                if isinstance(op, FieldNode):
                    yield GraphQLResult(data={target_from_node(op): event})

        else:
            raise NotImplementedError("Only subscription are mocked")



"""Rath

Rath is a GraphQL client for Python that is designed to be easy to use and
extendable. It follows the GraphQL spec and supports queries, mutations, and
subscriptions, with a focus on enabling apollo client like features in Python.

It can interface with a variety of transports and is designed to be easy to
extend and add new features to, thorugh the composition of links.

Rath is built on top of [pydantic](https://pydantic-docs.helpmanual.io/).

"""

from .rath import Rath

__all__ = ["Rath"]



from typing import AsyncIterator

from rath.links.base import ContinuationLink
from rath.operation import GraphQLResult, Operation
from rath.errors import NotComposedError


class ForwardLink(ContinuationLink):
    """The Forward link is a void link
    that simply forwards the operation to the next link.

    This link is useful for conditional fallbacks for other
    links.
    """

    async def aexecute(self, operation: Operation) -> AsyncIterator[GraphQLResult]:
        """Executes an operation against the link

        This link simply forwards the operation to the next link.


        Parameters
        ----------
        operation : Operation
            The operation to execute

        Yields
        ------
        GraphQLResult
            The result of the operation
        """
        if not self.next:
            raise NotComposedError("No next link set")

        async for result in self.next.aexecute(operation):
            yield result



from typing import Optional, Dict, Any, Union
from graphql.language import OperationDefinitionNode, print_ast
from graphql import (
    DocumentNode,
    get_operation_ast,
    parse,
)
from pydantic import BaseModel, ConfigDict, Field
import uuid


class Context(BaseModel):
    """Context provides a way to pass arbitrary data to resolvers on the context"""

    headers: Dict[str, str] = Field(default_factory=dict)
    files: Dict[str, Any] = Field(default_factory=dict)
    initial_payload: Dict[str, Any] = Field(default_factory=dict)
    kwargs: Dict[str, Any] = Field(default_factory=dict)


class Extensions(BaseModel):
    """Extensions is a map of additional metadata that can be used by the links in the chain"""

    pollInterval: Optional[int] = None
    maxPolls: Optional[int] = None


class Operation(BaseModel):
    """A GraphQL operation.

    An Operation is a GraphQL operation that can be executed by a GraphQL client.
    It is a combination of a query, variables, and headers, as well as a context
    that can be used to pass additional information to the link chain and
    extensions that can be used to pass additional information to the server.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    id: str = Field(default_factory=lambda: str(uuid.uuid4()))
    document_node: DocumentNode
    node: OperationDefinitionNode
    document: str
    variables: Dict[str, Any]
    operation_name: Optional[str]
    extensions: Extensions
    context: Context


class GraphQLResult(BaseModel):
    """GraphQLResult is the result of a GraphQL operation."""

    data: Dict[str, Any]


class GraphQLException(Exception):
    """GraphQLException is the base exception for all GraphQL errors."""

    pass


class SubscriptionDisconnect(GraphQLException):
    """SubscriptionDisconnect is raised when a subscription is disconnected."""

    pass


def opify(
    query: Union[str, DocumentNode],
    variables: Optional[Dict[str, Any]] = None,
    headers: Optional[Dict[str, Any]] = None,
    operation_name: Optional[str] = None,
    **kwargs,
) -> Operation:
    """Opify takes a query, variables, and headers and returns an Operation.

    Operations are the main way to interact with the GraphQL client, they are
    used to execute queries, mutations, and subscriptions, and can carry additional
    information in their context and extensions.

    Parameters
    ----------
    query : Union[str, DocumentNode]
        The query to execute
    variables : Optional[Dict[str, Any]], optional
        The variables to use, by default None
    headers : Optional[Dict[str, Any]], optional
        The headers to use, by default None
    operation_name : Optional[str], optional
        The operation name to use, by default None

    Returns
    -------
    Operation
        The operation that can be executed
    """

    document = parse(query) if isinstance(query, str) else query
    op = get_operation_ast(document, operation_name)
    assert op, f"No operation named {operation_name}"
    return Operation(
        node=op,
        document=print_ast(document),
        document_node=document,
        variables=variables or {},
        operation_name=operation_name,
        extensions=Extensions(),
        context=Context(headers=headers or {}, kwargs=kwargs),
    )



from typing import TypeVar

from koil.qt import QtRunner, QtFuture
from qtpy import QtCore
from rath import Rath
from turms.funcs import TurmsOperation
from pydantic import BaseModel

T = TypeVar("T")


class QtRathQuery(QtRunner):
    """QtRathQuery is a QtRunner that runs a query on a Rath instance.
    TODO: This should be more straightforward to use.
    """

    started: QtCore.Signal = QtCore.Signal()
    errored: QtCore.Signal = QtCore.Signal(Exception)
    cancelled: QtCore.Signal = QtCore.Signal()
    returned: QtCore.Signal = QtCore.Signal(object)

    def __init__(self, operation: TurmsOperation, rath: Rath) -> None:
        """Initializes the QtRathQuery

        Parameters
        ----------
        operation : Type[T]
            The operation to run
        rath : Rath
            The rath instance to run the operation on
        """

        async def coro(*args, **kwargs) -> BaseModel:
            """wrapper coroutine to run the operation on the rath instance"""
            assert rath is not None, "No rath found"

            result = await rath.aquery(operation.Meta.document, *args, **kwargs)
            return operation(**result.data)

        super().__init__(coro)
        self.operation = operation

    def run(self, **kwargs) -> QtFuture:
        """Runs the operation on the rath instance

        This is an async method that returns a QtFuture that can be used to cancel the operation.
        The result of the operation is emitted by the returned signal, and the error is emitted by the errored signal.

        Parameters
        ----------
        **kwargs
            The arguments to pass to the operation

        Returns
        -------
        QtFuture:
            The future of the operation, can be used to cancel the operation

        """

        arg = self.operation.Arguments(**kwargs).dict()  # type: ignore
        # TODO: This should be more straightforward to use, and typed
        return super().run(arg)



""" Herre contrib package

This package contains contrib modules for the herre package. This allows for the
creation of herre powered links that can be used to dynamially retrieve a token
using an herre grant (e.g. an interactive grant) and send it to the next link.
"""



from typing import Any, List


class NotQueriedError(Exception):
    """An error that is raised if a nested attribute is not queried"""

    pass


def get_nested_error(obj: Any, nested: List[str], above: List[str]) -> Any:
    """Get a nested attribute or raise an error


    Raises an error if a nested attribut is not present

    This is used to raise an error if a nested attribute is not present.
    And to give a hint where in a query tree a the nested attribute is missing.

    Parameters
    ----------
    obj : Any
        The object to query
    nested : List[str]
        The nested attributes to query
    above : _type_
        The attributes above the nested attribute ( will be filled by the function)

    Returns
    -------
    Any
        The queried object

    Raises
    ------
    NotQueriedError
        The nested attribute was not queried
    """
    if hasattr(obj, nested[0]):
        obj = getattr(obj, nested[0])
        if len(nested) > 1:
            return get_nested_error(obj, nested[1:], above + [nested[0]])
        else:
            return obj
    else:
        raise NotQueriedError(f"{nested} not queried. {above} was queried")


def get_attributes_or_error(object: Any, *args: str) -> Any:
    """Get attributes or raise an error

    Get attributes from an object or raise an error if the attribute is not present
    . Nested attributes can be queried by using a dot as a delimiter.

    Examples
    --------

    ```python
    @dataclass
    class B:
        c: int

    @dataclass
    class A:
        a: int
        b: B

    a = A(a=1, b=B(c=2))

    get_attributes_or_error(a, "a", "b.c") # (1, 2)
    ```

    Parameters
    ----------
    object : Any
        The object to query

    *args : str
        The attributes to query (nested attribtues delimited by a dot)

    Raises
    ------
    NotQueriedError
        The nested attribute was not queried






    """
    returns = []
    errors = []
    for i in args:
        if "." in i:
            try:
                returns.append(get_nested_error(object, i.split("."), []))
                continue
            except NotQueriedError:
                errors.append(i)
        else:
            if hasattr(object, i):
                returns.append(getattr(object, i))
            else:
                errors.append(i)

    if len(errors) > 0:
        raise NotQueriedError(
            f"Required fields {errors} not queried on {object.__class__.__name__}"
        )

    if len(args) == 1:
        return returns[0]
    else:
        return tuple(returns)



from rath.operation import Operation
from rath.links.parsing import ParsingLink


class ShrinkingLink(ParsingLink):
    """Shrinking Link is a link that shrinks operation variables.
    It traverses the variables dict, and converts any model that has a .ashrink() method to its id.
    """

    async def aparse(self, operation: Operation) -> Operation:
        """Parses an operation

        This method will traverse the variables dict, and convert any model that has a .ashrink() method to its id.
        This is a helper method that can be used to shrink the variables of an operation before sending it to the server.
        (Utilized in the arkitekt framework)
        """

        shrinked_variables = {
            key: await var.ashrink()
            for key, var in operation.variables.items()
            if hasattr(var, "ashrink")
        }
        operation.variables.update(shrinked_variables)
        return operation



""" Links module. 


The links module contains all the links that can be used to compose a client.
A link is a class that implements the Link interface. The Link interface is defined
in the base module. This module also contains some helper functions that can be used
to compose and combine links, to create more complex client behaviour.

"""

from .compose import compose
from .split import split

__all__ = ["compose", "split"]



from typing import AsyncIterator, Optional, cast, Dict, Any, Type
from graphql import (
    GraphQLSchema,
    build_ast_schema,
    build_client_schema,
    get_introspection_query,
    validate,
    IntrospectionQuery,
)
from graphql.language.parser import parse
from pydantic import model_validator
from rath.links.base import ContinuationLink
from rath.links.errors import ContinuationLinkError
from rath.operation import GraphQLResult, Operation, opify
from glob import glob


def schemify(
    schema_dsl: Optional[str] = None, schema_glob: Optional[str] = None
) -> GraphQLSchema:
    """Schemify creates a GraphQLSchema from a schema dsl or a glob to a set of graphql files

    Parameters
    ----------
    schema_dsl : Optional[str], optional
        The schema_dsl to use, by default None
    schema_glob : Optional[str], optional
        A path/glaob to the schema files, by default None

    Returns
    -------
    GraphQLSchema
        The schema

    """

    if schema_dsl:
        return build_ast_schema(parse(schema_dsl))
    if schema_glob:
        files = glob(schema_glob, recursive=True)
        dsl_string = ""
        for file in files:
            with open(file, "r") as f:
                if file.endswith(".graphql"):
                    dsl_string += f.read()

        assert dsl_string, f"No schema found in glob {schema_glob}"
        return build_ast_schema(parse(dsl_string))

    raise NotImplementedError("Please provide either a dsl or glob")


class ValidationError(ContinuationLinkError):
    """ValidationError is raised when a validation error occurs"""

    pass


class ValidatingLink(ContinuationLink):
    """ValidatingLink validates the operation againt as schema before passing it on to the next link.

    The schema can be provided as a dsl string, or as a glob to a set of graphql files.
    If the schema is not provided, the link will introspect the server to get the schema if allow_introspection is set to True.


    """

    schema_dsl: Optional[str] = None
    """ The schema (as a string) to validate against. If not provided, the link will introspect the server to get the schema if allow_introspection is set to True."""
    schema_glob: Optional[str] = None
    """ The glob to a set of graphql files to validate against. If not provided, the link will introspect the server to get the schema if allow_introspection is set to True."""
    allow_introspection: bool = False
    """ If set to True, the link will introspect the server to get the schema if it is not provided."""

    graphql_schema: Optional[GraphQLSchema] = None
    """ The schema to validate against. If not provided, the link will introspect the server to get the schema if allow_introspection is set to True."""

    @model_validator(mode="after")
    @classmethod
    def check_schema_dsl_or_schema_glob(cls: Type["ValidatingLink"], self: "ValidatingLink", *info) -> Dict[str, Any]:  # type: ignore
        """Validates and checks that either a schema_dsl or schema_glob is provided, or that allow_introspection is set to True"""
        if not self.schema_dsl and not self.schema_glob:
            if not self.allow_introspection:
                raise ValueError(
                    "Please provide either a schema_dsl or schema_glob or allow introspection"
                )

        else:
            self.graphql_schema = schemify(
                schema_dsl=self.schema_dsl,
                schema_glob=self.schema_glob,
            )

        return self

    async def introspect(self, starting_operation: Operation) -> GraphQLSchema:  # type: ignore
        """Introspects the server to get the schema

        Parameters
        ----------
        starting_operation : Operation
            The operation to use for introspection (can be used to set headers, etc.)

        Returns
        -------
        GraphQLSchema
            The introspected schema
        """
        if not self.next:
            raise ContinuationLinkError("No next link set")

        introspect_operation = opify(get_introspection_query())
        introspect_operation.context = starting_operation.context
        introspect_operation.extensions = starting_operation.extensions

        async for result in self.next.aexecute(introspect_operation):
            return build_client_schema(cast(IntrospectionQuery, result.data))

    async def aexecute(self, operation: Operation) -> AsyncIterator[GraphQLResult]:
        """Executes an operation against the link

        This link will validate the operation and then forward it to the next link,
        the next link will then execute the operation.

        Parameters
        ----------
        operation : Operation
            The operation to execute

        Yields
        ------
        GraphQLResult
            The result of the operation
        """
        if not self.next:
            raise ContinuationLinkError("No next link set")

        if not self.graphql_schema:
            assert self.allow_introspection, "Introspection is not allowed"
            self.graphql_schema = await self.introspect(operation)

        errors = validate(self.graphql_schema, operation.document_node)
        if len(errors) > 0:
            raise ValidationError(
                f"{operation} does not comply with the schema!\n Errors: \n\n"
                + "\n".join([e.message for e in errors])
            )

        async for result in self.next.aexecute(operation):
            yield result



from pydantic import BaseModel
from typing import Any, Type
from pydantic import GetCoreSchemaHandler
from pydantic_core import CoreSchema, core_schema

class ID(str):
    """A custom scalar for IDs. If passed a pydantic model it an id property
    it will automatically validate to that models id"""

    def __init__(self, value: str) -> None:
        """Initialize the ID"""
        self.value = value

    @classmethod
    def __get_pydantic_core_schema__(
        self,
        source_type: Any,  # noqa: ANN401
        handler: GetCoreSchemaHandler,  # noqa: ANN401
    ) -> CoreSchema:
        """Get the pydantic core schema for the interface"""
        return core_schema.no_info_before_validator_function(self.validate, handler(str))


    @classmethod
    def validate(cls: Type["ID"], v: Any) -> "ID":
        """Validate the ID"""
        if isinstance(v, BaseModel):
            if hasattr(v, "id"):
                return cls(v.id)  # type: ignore
            else:
                raise ValueError("It appears to be a pydantic model but has no id")

        if isinstance(v, str):
            return cls(v)

        if isinstance(v, int):
            return cls(str(v))

        raise TypeError(
            f"Needs to be either a instance of BaseModel (with an id) or a string got {type(v)}"
        )



from rath.links.base import AsyncTerminatingLink
from rath.operation import opify, GraphQLResult


async def run_basic_query(link: AsyncTerminatingLink) -> GraphQLResult:  # type: ignore
    """Run a basic query against the given link"""
    async with link:
        async for i in link.aexecute(
            operation=opify(
                """
                    query {
                        hello
                    }
                    """
            )
        ):
            return i



from rath.rath import Rath, current_rath
from typing import Optional, Dict, Any, Protocol, AsyncIterator, Iterator
from pydantic import BaseModel


class TurmsMeta(Protocol):
    """A TurmsMeta is a protocol that defines the meta class for a TurmsOperation"""

    document: str
    operation: str


class TurmsOperation(Protocol):
    """A TurmsOperation is a protocol that defines a GraphQL operation
    that was generated by Turms."""

    def __call__(self, *args: Any, **kwds: Any) -> BaseModel:
        """Mimics the __init__ method of base model"""
        ...

    @property
    def Meta(self) -> TurmsMeta:
        """A meta class that contains the document and operation name for the operation"""
        ...

    @property
    def __name__(self) -> str:
        """The name of the operation"""
        ...


def execute(
    operation: TurmsOperation, variables: Dict[str, Any], rath: Optional[Rath] = None
) -> BaseModel:
    """Synchronously Executes an a query or mutation using rath


    This function will execute an operation using rath, retrieving
    the currentliy active rath client from the current_rath context
    if no rath client is provided.

    Parameters
    ----------
    operation : TurmsOperation
        The turms operation to execute
    variables : Dict[str, Any]
        The variables to use
    rath : Optional[Rath], optional
        The rath client, by default the current rath client

    Returns
    -------
    BaseModel
        The result of the operation
    """
    rath = rath or current_rath.get()
    assert (
        rath is not None
    ), "No rath client provided and no rath client in current_rath context"
    return operation(**rath.query(operation.Meta.document, variables).data)


async def aexecute(
    operation: TurmsOperation, variables: Dict[str, Any], rath: Optional[Rath] = None
) -> BaseModel:
    """Asynchronously Executes a query or mutation using rath


    This function will execute an operation using rath, retrieving
    the currentliy active rath client from the current_rath context
    if no rath client is provided.

    Parameters
    ----------
    operation : TurmsOperation
        The turms operation to execute
    variables : Dict[str, Any]
        The variables to use
    rath : Optional[Rath], optional
        The rath client, by default the current rath client

    Returns
    -------
    BaseModel
        The result of the operation
    """
    rath = rath or current_rath.get()
    assert (
        rath is not None
    ), "No rath client provided and no rath client in current_rath context"
    x = await rath.aquery(operation.Meta.document, variables)
    return operation(**x.data)


def subscribe(
    operation: TurmsOperation, variables: Dict[str, Any], rath: Optional[Rath] = None
) -> Iterator[BaseModel]:
    """Synchronously subscribte to a subscription using rath


    This function will execute an operation using rath, retrieving
    the currentliy active rath client from the current_rath context
    if no rath client is provided.

    Parameters
    ----------
    operation : TurmsOperation
        The turms operation to execute
    variables : Dict[str, Any]
        The variables to use
    rath : Optional[Rath], optional
        The rath client, by default the current rath client

    Yields
    -------
    BaseModel
        The result of the operation
    """
    rath = rath or current_rath.get()
    assert (
        rath is not None
    ), "No rath client provided and no rath client in current_rath context"

    for event in rath.subscribe(operation.Meta.document, variables):
        yield operation(**event.data)


async def asubscribe(
    operation: TurmsOperation, variables: Dict[str, Any], rath: Optional[Rath] = None
) -> AsyncIterator[BaseModel]:
    """Asynchronously subscribte to a subscription using rath


    This function will execute an operation using rath, retrieving
    the currentliy active rath client from the current_rath context
    if no rath client is provided.

    Parameters
    ----------
    operation : TurmsOperation
        The turms operation to execute
    variables : Dict[str, Any]
        The variables to use
    rath : Optional[Rath], optional
        The rath client, by default the current rath client

    yields
    -------
    BaseModel
        The result of the operation
    """
    rath = rath or current_rath.get()
    assert (
        rath is not None
    ), "No rath client provided and no rath client in current_rath context"

    async for event in rath.asubscribe(operation.Meta.document, variables):
        yield operation(**event.data)



from typing import AsyncIterator, Optional
from koil.composition import KoiledModel
from rath.operation import GraphQLResult, Operation
from rath.errors import NotComposedError


class Link(KoiledModel):
    """A Link is a class that can be used to send operations to a GraphQL API.
    its main method is aexecute, which takes an operation and returns an
    AsyncIterator of GraphQLResults.

    Links can be composed to form a chain of links. The last link in the chain
    is the terminating link, which is responsible for sending the operation to
    the server.

    """

    async def aconnect(self, operation: Operation) -> None:
        """A coroutine that is called when the link is connected."""
        pass

    async def adisconnect(self) -> None:
        """A coroutine that is called when the link is disconnected."""
        pass

    async def __aenter__(self) -> None:
        """A coroutine that is called when the link is entered."""
        pass

    async def __aexit__(self, *args, **kwargs) -> None:
        """A coroutine that is called when the link is exited."""
        pass

    def aexecute(self, operation: Operation) -> AsyncIterator[GraphQLResult]:
        """A coroutine that takes an operation and returns an AsyncIterator
        of GraphQLResults. This method should be implemented by subclasses."""
        raise NotImplementedError(
            f"Please overwrite the asubscribe method in {self.__class__.__name__}"
        )


class TerminatingLink(Link):
    """TerminatingLink

    This link type is responsible for sending the operation to the server.

    The last link in a link chain MUST always a TerminatingLink. It cannot delegate the operation
    to another link.
    """


class AsyncTerminatingLink(TerminatingLink):
    """An Async Termination Link

    This is a link that terminates the query or subscription, aka
    it is a link that does not have a next link. It normally
    is used to make network requests with the already parsed
    operations

    Args:
        TerminatingLink (_type_): _description_
    """

    def aexecute(self, operation: Operation) -> AsyncIterator[GraphQLResult]:
        """Executes an operation against the link

        This is the main method of the link. It takes an operation and returns an AsyncIterator
        of GraphQLResults. This method should be implemented by subclasses. It is important
        that even if the operation is a query or mutation, this method should return an AsyncIterator
        (even if it only yields one result).

        Parameters
        ----------
        operation : Operation
            The operation to execute

        Yields
        ------
        GraphQLResult
            The result of the operation

        """

        raise NotImplementedError("Your Async Transport needs to overwrite this method")


class ContinuationLink(Link):
    """A ContinuationLink is a link that delegates the operation to the next link in the chain.
    It can be either provided a next link or it can be set once the link is composed together
    """

    next: Optional[Link] = None
    """The next link in the chain. This is also set when the link is composed together."""

    def set_next(self, next: Link) -> None:
        """Compose the link with the next link in the chain

        This method is called when the link is composed together with the next link in the chain.
        by the compose helper, this ensures that you cna use the "next" method to delegate the operation
        to the next link in the chain.

        Parameters
        ----------
        next : Link
            The next link in the chain
        """
        self.next = next

    async def aexecute(self, operation: Operation) -> AsyncIterator[GraphQLResult]:
        """Executes an operation against the NEXT link

        This is the main method of the link. It takes an operation and returns an AsyncIterator
        of GraphQLResults. This method should be implemented by subclasses. It is important
        To note that a ContinuationLink should always delegate the operation to the next link
        in the chain. This means that the next link should be called with the same operation.

        Parameters
        ----------
        operation : Operation
            The operation to execute

        Yields
        ------
        GraphQLResult
            The result of the operation

        """
        if not self.next:
            raise NotComposedError("No next link set")

        async for x in self.next.aexecute(operation):
            yield x



from ssl import SSLContext
from typing import AsyncIterator, Dict, Optional, Any
from graphql import OperationType
from pydantic import Field
import websockets
import json
import asyncio
import logging
import ssl
import certifi
from rath.links.errors import LinkNotConnectedError, TerminatingLinkError


from rath.operation import (
    GraphQLException,
    GraphQLResult,
    Operation,
    SubscriptionDisconnect,
)
from rath.links.base import AsyncTerminatingLink


logger = logging.getLogger(__name__)


GQL_WS_SUBPROTOCOL = "graphql-ws"

# all the message types
GQL_CONNECTION_INIT = "connection_init"
GQL_START = "start"
GQL_STOP = "stop"
GQL_CONNECTION_TERMINATE = "connection_terminate"
GQL_CONNECTION_ERROR = "connection_error"
GQL_CONNECTION_ACK = "connection_ack"
GQL_DATA = "data"
GQL_ERROR = "error"
GQL_COMPLETE = "complete"
GQL_CONNECTION_KEEP_ALIVE = "ka"

WEBSOCKET_DEAD = "websocket_dead"
WEBSOCKET_CANCELLED = "websocket_cancelled"


class CorrectableConnectionFail(TerminatingLinkError):
    """Raied when the connection fails, but can be recovered from"""

    pass


class DefiniteConnectionFail(TerminatingLinkError):
    """Raised when the connection fails, and cannot be recovered from"""

    pass


class InvalidPayload(TerminatingLinkError):
    """Raised when the payload is invalid"""

    pass


async def none_token_loader() -> None:
    """A token loader that always excepts"""
    raise Exception("Token loader was not set")


class SubscriptionTransportWsLink(AsyncTerminatingLink):
    """WebSocketLink is a terminating link that sends operations over websockets using
      websockets via the subscription-transport-ws protocol. This is a
      deprecated protocol, and should not be used for new projects.

    This is a terminating link, so it should be the last link in the chain.
    This is a stateful link, keeing a connection open and sending messages over it.

    """

    ws_endpoint_url: str
    """ The endpoint url to connect to """
    allow_reconnect: bool = True
    """ Should the websocket try to reconnect if it fails """
    time_between_retries: float = 4
    """ The sleep time between retries """
    max_retries: int = 3
    """ The maximum amount of retries before giving up """
    ssl_context: SSLContext = Field(
        default_factory=lambda: ssl.create_default_context(cafile=certifi.where())
    )
    """ The SSL Context to use for the connection """
    payload_token_to_querystring: bool = True
    """Should the payload token be sent as a querystring instead (as connection params
      is not supported by all servers)"""

    _connection_lock: Optional[asyncio.Lock] = None
    _connected: bool = False
    _alive: bool = False
    _send_queue: Optional[asyncio.Queue] = None
    _connection_task: Optional[asyncio.Task] = None
    _ongoing_subscriptions: Optional[Dict[str, asyncio.Queue]] = None

    async def aforward(self, message: str) -> None:
        """Forwards a message to the websocket

        Parameters
        ----------
        message : str
            The message to forward

        Raises
        ------
        LinkNotConnectedError
            Raised if the link is not connected
        """
        if not self._send_queue:
            raise LinkNotConnectedError("Link is not connected")
        await self._send_queue.put(message)

    async def __aenter__(self) -> None:
        """Enters the context manager of the link"""
        self._ongoing_subscriptions = {}
        self._send_queue = asyncio.Queue()
        self._connection_lock = asyncio.Lock()

    async def aconnect(self, initiating_operation: Operation) -> None:
        """Connects the websocket

        This method is called automatically when the link is entered.
        It should not be called manually. It needs to received
        an initiating operation, which is used to build the connection
        payload. It succeds when the connection is established, and fails
        if the connection cannot be established.

        Parameters
        ----------
        initiating_operation : Operation
            The initiating operation

        """
        logger.info("Connecting Websockets")
        connection_future = asyncio.get_running_loop().create_future()
        self._connection_task = asyncio.create_task(
            self.websocket_loop(initiating_operation, connection_future)
        )
        await connection_future

    async def adisconnect(self) -> None:
        """Disconnects the websocket"""
        if self._connection_task:
            self._connection_task.cancel()

            try:
                await self._connection_task
            except asyncio.CancelledError:
                logger.info(f"Websocket Transport {self} succesfully disconnected")

    async def __aexit__(self, *args, **kwargs) -> None:
        """Exits the context manager of the link"""
        await self.adisconnect()

    async def build_url(self, initiating_operation: Operation) -> str:
        """Builds the url for the websocket

        It uses the initiating operation to build the url. By default
        it will use the payload token as a querystring, but this can be
        disabled by setting payload_token_to_querystring to False.

        Parameters
        ----------
        initiating_operation : Operation
            The initiating operation

        Returns
        -------
        str
            The url for the websocket
        """
        if self.payload_token_to_querystring:
            token = initiating_operation.context.initial_payload.get("token", None)
            return (
                f"{self.ws_endpoint_url}?token={token}"
                if token
                else self.ws_endpoint_url
            )
        else:
            return self.ws_endpoint_url

    async def websocket_loop(
        self,
        initiating_operation: Operation,
        connection_future: asyncio.Future,
        retry: int = 0,
    ) -> None:
        """The main websocket loop

        This method is the main websocket loop. It will try to connect to the
        websocket, and will retry if it fails. It will also try to reconnect
        if the connection is lost.
        You should not call this method manually.
        """
        send_task = None
        receive_task = None
        try:
            try:
                url = await self.build_url(initiating_operation)
                async with websockets.connect(  # type: ignore
                    url,
                    subprotocols=[GQL_WS_SUBPROTOCOL],
                    ssl=self.ssl_context if url.startswith("wss") else None,
                ) as client:
                    logger.info("Websocket successfully connected")

                    send_task = asyncio.create_task(
                        self.sending(client, initiating_operation)
                    )
                    receive_task = asyncio.create_task(
                        self.receiving(client, initiating_operation, connection_future)
                    )

                    self._alive = True
                    done, pending = await asyncio.wait(
                        [send_task, receive_task],
                        return_when=asyncio.FIRST_EXCEPTION,
                    )
                    self._alive = False

                    for task in pending:
                        task.cancel()

                    for task in done:
                        exception = task.exception()
                        if exception:
                            raise exception
                        else:
                            raise CorrectableConnectionFail(
                                f"Websocket connection closed without exception: This is unexpected behaviours. Results ist {task.result()}"
                            )

            except Exception as e:
                logger.warning("Websocket excepted. Trying to recover", exc_info=True)
                raise CorrectableConnectionFail from e

        except CorrectableConnectionFail as e:
            logger.info(
                f"Trying to Recover from Exception {e} Reconnect is {self.allow_reconnect} Retry: {retry}"
            )
            if retry > self.max_retries or not self.allow_reconnect:
                logger.error("Max retries reached. Aborting")
                raise DefiniteConnectionFail("Exceeded Number of Retries")

            await asyncio.sleep(self.time_between_retries)
            logger.info("Retrying to connect")
            await self.broadcast(
                {"type": WEBSOCKET_DEAD, "error": e}, connection_future
            )
            await self.websocket_loop(
                initiating_operation, connection_future, retry=retry + 1
            )

        except DefiniteConnectionFail as e:
            logger.error("Websocket excepted closed definetely", exc_info=True)
            self.connection_dead = True
            if connection_future and not connection_future.done():
                connection_future.set_exception(e)
            raise e

        except asyncio.CancelledError as e:
            logger.info("Websocket got cancelled. Trying to shutdown graceully")
            if send_task and receive_task:
                send_task.cancel()
                receive_task.cancel()

                await asyncio.gather(send_task, receive_task, return_exceptions=True)
            raise e

        except Exception as e:
            logger.error("Websocket excepted", exc_info=True)
            self.connection_dead = True
            if connection_future and not connection_future.done():
                connection_future.set_exception(e)
            raise e

    async def sending(self, client: Any, initiating_operation: Operation) -> None:
        """The sending task

        This method is the sending task. It will send messages to the websocket
        as they are put into the send queue.  It should not be called manually.
        but is called automatically by the websocket loop.


        Parameters
        ----------
        client : websocket.Client
            The websockets client
        initiating_operation : Operation
            The initiating operation

        """
        payload = {
            "type": GQL_CONNECTION_INIT,
            "payload": {"headers": initiating_operation.context.headers},
        }
        await client.send(json.dumps(payload))

        try:
            while True:
                if not self._send_queue:
                    raise LinkNotConnectedError("Link is not connected")

                message = await self._send_queue.get()
                logger.debug("GraphQL Websocket: >>>>>> " + message)
                await client.send(message)
                self._send_queue.task_done()
        except asyncio.CancelledError as e:
            logger.debug("Sending Task sucessfully Cancelled")  #
            raise e

    async def receiving(
        self,
        client: Any,
        initiating_operation: Operation,
        connection_future: asyncio.Future,
    ) -> None:
        """The receiving task

        This method is the receiving task. It will receive messages from the websocket
        and broadcast them to the subscriptions. It should not be called manually.
        but is called automatically by the websocket loop.

        Parameters
        ----------
        client : websocket.Client
            The websockets client
        initiating_operation : Operation
            The initiating operation
        connection_future : asyncio.Future
            The future to set when the connection is established



        """
        try:
            async for message in client:
                logger.debug("GraphQL Websocket: <<<<<<< " + message)
                try:
                    message = json.loads(message)
                    await self.broadcast(message, connection_future)
                except json.JSONDecodeError as err:
                    logger.warning(
                        "Ignoring. Server sent invalid JSON data: %s \n %s",
                        message,
                        err,
                    )
        except Exception as e:
            logger.warning("Websocket excepted. Trying to recover", exc_info=True)
            raise e

    async def broadcast(self, message: dict, connection_future: asyncio.Future) -> None:
        """Broadcasts a message to the subscriptions, or handles it internally
        e.g if it is a connection ack message.

        Parameters
        ----------
        message : dict
            The message to broadcast ( already json parsed)
        connection_future : asyncio.Future
            The future to set when the connection is established
        """
        type = message["type"]

        if type == GQL_CONNECTION_ACK:
            if connection_future and not connection_future.done():
                connection_future.set_result(True)
            return

        if type == GQL_CONNECTION_KEEP_ALIVE:
            if connection_future and not connection_future.done():
                connection_future.set_result(True)
            return

        if type == WEBSOCKET_DEAD:
            # notify all subscriptipns that the websocket is dead
            if not self._ongoing_subscriptions:
                self._ongoing_subscriptions = {}

            for subscription in self._ongoing_subscriptions.values():
                await subscription.put(message)
            return

        if type in [GQL_DATA, GQL_COMPLETE]:
            if "id" not in message:
                raise InvalidPayload(f"Protocol Violation. Expected 'id' in {message}")

            id = message["id"]
            if not self._ongoing_subscriptions:
                self._ongoing_subscriptions = {}

            assert (
                id in self._ongoing_subscriptions
            ), "Received Result for subscription that is no longer or was never active"
            await self._ongoing_subscriptions[id].put(message)

    async def aexecute(self, operation: Operation) -> AsyncIterator[GraphQLResult]:
        """Executes an operation against the link

        This link will send the operation to the websocket, and then
        wait for the result.

        Parameters
        ----------
        operation : Operation
            The operation to execute

        Yields
        ------
        GraphQLResult
            The result of the operation
        """
        if not self._connection_lock:
            raise Exception(
                "WebsocketLink not entered yet. Please us this in an async context manager"
            )

        async with self._connection_lock:
            if self._connection_task is None or self._connection_task.done():
                await self.aconnect(operation)

        assert (
            operation.node.operation == OperationType.SUBSCRIPTION
        ), "Operation is not a subscription"
        assert not operation.context.files, "We cannot send files through websockets"

        id = operation.id
        subscribe_queue = asyncio.Queue()  # type: asyncio.Queue

        if not self._ongoing_subscriptions:
            self._ongoing_subscriptions = {}

        self._ongoing_subscriptions[id] = subscribe_queue

        send_payload = {
            "headers": operation.context.headers,
            "query": operation.document,
            "variables": operation.variables,
        }

        try:
            frame = {"id": id, "type": GQL_START, "payload": send_payload}
            await self.aforward(json.dumps(frame))
            logger.debug(f"Subcription started {operation}")

            while True:
                answer = await subscribe_queue.get()

                if answer["type"] == GQL_DATA:
                    payload = answer["payload"]

                    if "errors" in payload:
                        raise GraphQLException(
                            "\n".join([e["message"] for e in payload["errors"]])
                        )

                    if "data" in payload:
                        yield GraphQLResult(data=payload["data"])
                        subscribe_queue.task_done()

                if answer["type"] == WEBSOCKET_DEAD:
                    raise SubscriptionDisconnect(
                        f"Subcription {id} failed propagating Error {operation}"
                    )

                if answer["type"] == GQL_COMPLETE:
                    logger.info(f"Subcription done {operation}")
                    return

        except asyncio.CancelledError as e:
            logger.debug(f"Subcription ended {operation}")
            await self.aforward(json.dumps({"id": id, "type": GQL_STOP}))
            raise e



from rath.links.base import ContinuationLink
from rath.operation import GraphQLResult, Operation
from typing import AsyncIterator
from rath.errors import NotComposedError


class ParsingLink(ContinuationLink):
    """ParsingLink is a link that parses operation and returns a new operation.
    It is an abstract class that needs to be implemented by the user.
    """

    async def aparse(self, operation: Operation) -> Operation:
        """Parses an operation

        This method needs to be implemented by the user.
        It should parse the operation and return a new operation.

        Parameters
        ----------
        operation : Operation
            The operation to execute

        Returns
        -------
        Operation
            The parsed operation
        """
        raise NotImplementedError("Please implement this method")

    async def aexecute(
        self, operation: Operation, **kwargs
    ) -> AsyncIterator[GraphQLResult]:
        """Executes an operation against the link

        This link will parse the operation and then forward it to the next link,
        the next link will then execute the operation.

        Parameters
        ----------
        operation : Operation
            The operation to execute

        Yields
        ------
        GraphQLResult
            The result of the operation
        """
        if not self.next:
            raise NotComposedError("No next link set")

        operation = await self.aparse(operation)
        async for result in self.next.aexecute(operation, **kwargs):
            yield result



from typing import AsyncIterator, List, Optional, Type, Any

from pydantic import field_validator
from rath.links.base import ContinuationLink, Link, TerminatingLink
from rath.operation import GraphQLResult, Operation
from rath.errors import NotComposedError


class ComposedLink(TerminatingLink):
    """A composed link is a link that is composed of multiple links. The links
    are executed in the order they are passed to the constructor.

    The first link in the chain is the first link that is executed. The last
    link in the chain is the terminating link, which is responsible for sending
    the operation to the server.
    """

    links: List[Link]
    """The links that are composed to form the chain. pydantic will validate 
    that the last link is a terminating link."""

    @field_validator("links")
    def validate(cls: Type["ComposedLink"], value: Any) -> List[Link]:
        """Validate that the links are valid"""
        if not value:
            raise ValueError("ComposedLink requires at least one link")

        if not all(isinstance(link, Link) for link in value):
            raise ValueError("ComposedLink requires all links to be Links")

        assert isinstance(
            value[-1], TerminatingLink
        ), "Last link must be a TerminatingLink"
        for link in value[:-1]:
            assert isinstance(
                link, ContinuationLink
            ), f"All but the last must be ContinuationLinks: check {link}"

        return value

    async def aconnect(self, operation: Operation) -> None:
        """Delegate the connect to all the links"""
        for link in self.links:
            await link.aconnect(operation)

    async def adisconnect(self) -> None:
        """Delegate the disconnect to all the links"""
        for link in self.links:
            await link.adisconnect()

    async def __aenter__(self) -> None:
        """Compose the links together and set the first link as the first link"""
        # We need to make sure that the links are connected in the correct order
        for i in range(len(self.links) - 1):
            self.links[i].set_next(self.links[i + 1])

        for link in self.links:
            await link.__aenter__()

    async def __aexit__(self, *args, **kwargs) -> None:
        """Exit the links in reverse order"""
        for link in self.links:
            await link.__aexit__(*args, **kwargs)

    async def aexecute(self, operation: Operation) -> AsyncIterator[GraphQLResult]:
        """Execute the first link

        This is the main method of the link. It takes an operation and returns an AsyncIterator
        of GraphQLResults. This method should be implemented by subclasses. It is important
        """
        async for result in self.links[0].aexecute(operation):
            yield result


class TypedComposedLink(TerminatingLink):
    """A typed composed link is a base class to create a definition for
    a composed link. It is a link that is composed of multiple links that
    will be set at compile time and not at runtime. Just add links
    that you want to use to the class definition and they will be
    automatically composed together.
    """

    _firstlink: Optional[Link] = None

    async def __aenter__(self) -> None:
        """Compose the links together and set the first link as the first link"""

        current_link = None

        for key, link in self:
            if isinstance(link, Link):
                if current_link:
                    current_link.set_next(link)
                    await current_link.__aenter__()
                else:
                    self._firstlink = link
                    await self._firstlink.__aenter__()

                current_link = link

        if current_link:
            await current_link.__aenter__()
        else:
            raise NotComposedError("No links set")

    async def __aexit__(self, *args, **kwargs) -> None:
        """Exit the links in reverse order"""
        for key, link in self:
            if isinstance(link, Link):
                await link.__aexit__(*args, **kwargs)

    async def aexecute(self, operation: Operation) -> AsyncIterator[GraphQLResult]:
        """Execute the first link"""
        if not self._firstlink:
            raise NotComposedError(
                "Links need to be composed before they can be executed. (Through __aenter__)"
            )

        async for result in self._firstlink.aexecute(operation):
            yield result


def compose(*links: Link) -> ComposedLink:
    """Compose a chain of links together. The first link in the chain is the first link that is executed.
    The last link in the chain is the terminating link, which is responsible for sending the operation to the server.


    Returns:
        ComposedLink: The composed link
    """

    return ComposedLink(links=links)



""" Testing package for rath.links


This package contains links that are useful for testing.
Like AssertLink, which asserts that the operation is valid.
Or MockLink, which mocks the response of an operation.

"""



""" Fakts contrib package

This package contains contrib modules for the fakts package.
It provides links that can be to connect to a graphql-server
whose endpoint is not known at "compile" time. (e.g. a server
whose ip-adress is dynamic). These links use the fakts protocol
to automatically discover the graphql endpoint.

"""



class RathException(Exception):
    """RathException is the base exception for all Rath errors."""

    pass


class NotConnectedError(RathException):
    """NotConnectedError is raised when the Rath is not connected and autoload
    is set to false."""

    pass


class NotEnteredError(RathException):
    """NotEnteredError is raised when the Rath is not entered and access
    to protected methods is attempted."""

    pass


class NotComposedError(RathException):
    """NotComposedError is raised when the Rath link chain is not composed and
    the next link is accessed."""

    pass



from typing import AsyncIterator, Optional

from rath.links.base import ContinuationLink
from rath.operation import (
    GraphQLException,
    GraphQLResult,
    Operation,
    SubscriptionDisconnect,
)
import logging
import asyncio
from rath.errors import NotComposedError

logger = logging.getLogger(__name__)


class RetryLink(ContinuationLink):
    """RetriyLink is a link that retries a operation  fails.
    This link is stateful, and will keep track of the number of times the
    subscription has been retried."""

    maximum_retry_attempts: int = 3
    """The maximum number of times the operation function will be called, before the operation fails."""
    sleep_interval: Optional[int] = None
    """The number of seconds to wait before retrying the operation."""

    async def aexecute(
        self, operation: Operation, retry: int = 0
    ) -> AsyncIterator[GraphQLResult]:
        """Executes an operation against the link

        This link will retry the operation if it fails.
        It will retry the operation a maximum of maximum_retry_attempts times.
        If a sleep_interval is set, it will wait that many seconds before retrying.

        Parameters
        ----------
        operation : Operation
            The operation to execute

        Yields
        ------
        GraphQLResult
            The result of the operation
        """

        if not self.next:
            raise NotComposedError("No next link set")

        try:
            async for result in self.next.aexecute(operation):
                yield result

        except SubscriptionDisconnect as e:
            if retry > self.maximum_retry_attempts:
                raise GraphQLException(
                    "Maximum refresh attempts reached. Trying to rescribe to"
                ) from e
            if self.sleep_interval:
                await asyncio.sleep(self.sleep_interval)

            logger.info(f"Subscription {operation} disconnected. Retrying {retry}")
            async for result in self.aexecute(operation, retry=retry + 1):
                yield result



from typing import Any, Dict, Optional

from pydantic import BaseModel
from fakts.fakts import Fakts
from rath.links.graphql_ws import GraphQLWSLink


class WebsocketHttpConfig(BaseModel):
    """A WebsocketHttpConfig is a Fakt that can be used to configure the aiohttp client."""

    ws_endpoint_url: str


class FaktsGraphQLWSLink(GraphQLWSLink):
    """FaktsGraphQLWSLink


    A FaktsGraphQLWSLink is a GraphQLWSLink that retrieves the configuration
    from a passed fakts context.

    """

    fakts: Fakts
    """The fakts context to use for configuration"""
    ws_endpoint_url: Optional[str]  # type: ignore
    fakts_group: str = "websocket"
    """ The fakts group within the fakts context to use for configuration """

    _old_fakt: Dict[str, Any] = {}

    def configure(self, fakt: WebsocketHttpConfig) -> None:
        """Configure the link with the given fakt"""
        self.ws_endpoint_url = fakt.ws_endpoint_url

    async def aconnect(self, operation: Any) -> None:
        """Connects the link to the server

        This method will retrieve the configuration from the fakts context,
        and configure the link with it. Before connecting, it will check if the
        configuration has changed, and if so, it will reconfigure the link.
        """
        if self.fakts.has_changed(self._old_fakt, self.fakts_group):
            self._old_fakt = await self.fakts.aget(self.fakts_group)
            assert self._old_fakt is not None, "Fakt should not be None"
            self.configure(WebsocketHttpConfig(**self._old_fakt))

        return await super().aconnect(operation)



from herre.herre import Herre
from rath.links.auth import AuthTokenLink
from rath.operation import Operation


class HerreAuthLink(AuthTokenLink):
    """HerreAuthLink is a link that retrieves a token from herre and sends it to the next link."""

    herre: Herre

    async def aload_token(self, operation: Operation) -> str:
        """Retrieves the token from herre"""
        herre = self.herre
        return await herre.aget_token()

    async def arefresh_token(self, operation: Operation) -> str:
        """Refreshes the token from herre"""
        herre = self.herre
        return await herre.arefresh_token()



from rath.errors import RathException


class LinkError(RathException):
    """Base class for all link errors."""


class LinkNotConnectedError(LinkError):
    """LinkNotConnectedError is raised when the link is not connected and autoload is set to false."""

    pass

    def __init__(self, message: str) -> None:
        """_summary_

        Parameters
        ----------
        message : _type_
            The message to raise (will be preprended to the default message)
        """
        super().__init__(
            message
            + "\n To connect link please use either an async or sync context manager "
        )


class TerminatingLinkError(LinkError):
    """Raised when a terminating link is called.

    This is a base class for all terminating link errors."""


class ContinuationLinkError(LinkError):
    """Raised when a continuation link is called an errors.

    THis is a base class for all continuation link errors."""

    pass


class AuthenticationError(TerminatingLinkError):
    """Signals that the authentication failed."""

    pass



from typing import AsyncIterator, Callable, List
from rath.links.base import AsyncTerminatingLink
from rath.operation import GraphQLResult, Operation


class AssertLink(AsyncTerminatingLink):
    """AssertLink is a link that asserts that the operation matches a set of assertions.

    This link is useful for testing.

    """

    assertions: List[Callable[[Operation], bool]] = []

    async def aexecute(self, operation: Operation) -> AsyncIterator[GraphQLResult]:
        """Executes an operation against the link

        Parameters
        ----------
        operation : Operation
            The operation to execute

        Yields
        ------
        GraphQLResult
            The result of the operation
        """
        for assertion in self.assertions:
            assert assertion(
                operation
            ), f"Assertion failed for operation {operation} on {assertion}"

        yield GraphQLResult(data={})



from datetime import datetime
from http import HTTPStatus
import json
from ssl import SSLContext
from typing import Any, Dict, List, Type, AsyncIterator
from rath.links.types import Payload
import aiohttp
from graphql import OperationType
from pydantic import Field
from rath.operation import GraphQLException, GraphQLResult, Operation
from rath.links.base import AsyncTerminatingLink
from rath.links.errors import AuthenticationError
import logging
import certifi
import ssl

logger = logging.getLogger(__name__)


class DateTimeEncoder(json.JSONEncoder):
    """DateTimeEncoder is a JSONEncoder that extends the default python json decoder to serialize"""

    def default(self, o):  # noqa
        """Override the default method to serialize datetime objects to ISO 8601 strings"""
        if isinstance(o, datetime):
            return o.isoformat()

        return json.JSONEncoder.default(self, o)


class AIOHttpLink(AsyncTerminatingLink):
    """AIOHttpLink is a terminating link that sends operations over HTTP using aiohttp.

    Aiohttp is a Python library for asynchronous HTTP requests. This link uses the
    standard aiohttp library to send operations over HTTP, but provides an ssl context
    that is configured to use the certifi CA bundle by default. You can override this
    behavior by passing your own SSLContext to the constructor.
    """

    endpoint_url: str
    """endpoint_url is the URL to send operations to."""
    ssl_context: SSLContext = Field(
        default_factory=lambda: ssl.create_default_context(cafile=certifi.where())
    )
    """ssl_context is the SSLContext to use for the aiohttp session. By default, this
    is a context that uses the certifi CA bundle."""

    auth_errors: List[HTTPStatus] = Field(
        default_factory=lambda: (HTTPStatus.FORBIDDEN,)
    )
    """auth_errors is a list of HTTPStatus codes that indicate that the request was
    unauthorized. By default, this is just HTTPStatus.FORBIDDEN, but you can
    override this to include other status codes that indicate that the request was
    unauthorized."""

    json_encoder: Type[json.JSONEncoder] = Field(default=DateTimeEncoder, exclude=True)
    """json_encoder is the JSONEncoder to use when serializing the payload. By default,
    this is a DateTimeEncoder that extends the default python json decoder to serialize 
    datetime objects to ISO 8601 strings."""

    _connected = False

    async def __aenter__(self) -> None:
        """Entery point for the async context manager"""
        pass

    async def aconnect(self, operation: Operation) -> None:
        """Connects the link to the server

        In the standard aiohttp transport, this method does nothing, as the
        connection is established when the request is sent.

        """
        self._connected = True

    async def __aexit__(self, *args, **kwargs) -> None:
        """Exit point for the async context manager"""
        pass

    async def aexecute(self, operation: Operation) -> AsyncIterator[GraphQLResult]:
        """Executes an operation against the link

        This link will create a new aiohttp session for each request. While
        we could also reuse the session, we currently don't do this. If
        you feel like this should be changed, please open an issue.

        Parameters
        ----------
        operation : Operation
            The operation to execute

        Yields
        ------
        GraphQLResult
            The result of the operation
        """
        if not self._connected:
            await self.aconnect(operation)

        payload: Payload = {"query": operation.document}

        if operation.node.operation == OperationType.SUBSCRIPTION:
            raise NotImplementedError(
                "Aiohttp Transport does not support subscriptions"
            )

        if len(operation.context.files.items()) > 0:
            payload["variables"] = operation.variables

            files = operation.context.files
            data = aiohttp.FormData()

            file_map = {str(i): [path] for i, path in enumerate(files)}

            # Enumerate the file streams
            # Will generate something like {'0': <_io.BufferedReader ...>}
            file_streams = {str(i): files[path] for i, path in enumerate(files)}
            operations_str = json.dumps(payload, cls=self.json_encoder)

            data.add_field(
                "operations", operations_str, content_type="application/json"
            )
            file_map_str = json.dumps(file_map)
            data.add_field("map", file_map_str, content_type="application/json")

            for k, v in file_streams.items():
                data.add_field(
                    k,
                    v,
                    filename=getattr(v, "name", k),
                )

            post_kwargs: Dict[str, Any] = {"data": data}

        else:
            payload["variables"] = operation.variables
            post_kwargs = {"json": payload}

        async with aiohttp.ClientSession(
            connector=aiohttp.TCPConnector(ssl=self.ssl_context),
            json_serialize=lambda x: json.dumps(x, cls=self.json_encoder),
        ) as session:
            async with session.post(
                self.endpoint_url, headers=operation.context.headers, **post_kwargs
            ) as response:
                if response.status == HTTPStatus.OK:
                    await response.json()

                if response.status in self.auth_errors:
                    raise AuthenticationError(
                        f"Token Expired Error {operation.context.headers}"
                    )

                json_response = await response.json()

                if "errors" in json_response:
                    raise GraphQLException(
                        "\n".join([e["message"] for e in json_response["errors"]])
                    )

                if "data" not in json_response:
                    raise Exception(f"Response does not contain data {json_response}")

                yield GraphQLResult(data=json_response["data"])



""" THe Qt Rath package.

This package contains the QtRathQuery class, which is a subclass of RathQuery
that emits Qt signals when the operation is started, errored, cancelled, or
returned.  This allows you to use it in a Qt application without blocking the
main thread. Attention: This is a work in progress, and is not yet ready for
use.

TODO: This should be more straightforward to use.

"""

from .helpers import QtRathQuery

__all__ = ["QtRathQuery"]



""" Fakts contrib package

This package contains contrib modules for the fakts package. In this case, it
contains links that can be used to connect to a graphql-server whose endpoint
is not known at "compile" time. (e.g. a server whose ip-adress is dynamic).
These link uses the fakts protocol to automatically discover the graphql endpoint.
"""
